# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Global fixtures for pytest."""
import collections.abc as cabc
import importlib.metadata as imm
import io
import pathlib
import sys

import pytest

import capellambse

INSTALLED_PACKAGE = pathlib.Path(capellambse.__file__).parent

TEST_ROOT = pathlib.Path(__file__).parent / "data" / "melodymodel"
TEST_MODEL = "Melody Model Test.aird"

capellambse.load_model_extensions()

if not any(
    i.module.startswith("capellambse.extensions.")
    for i in imm.entry_points(group="capellambse.model_extensions")
):
    raise RuntimeError(
        "Built-in model extensions are not loaded,"
        " is capellambse installed properly?"
    )


@pytest.fixture(scope="session")
def session_shared_model() -> cabc.Iterator[capellambse.MelodyModel]:
    """Load the standard test model.

    Unlike the ``model`` fixture, this fixture is shared across the
    entire test session. As such, the test functions using this fixture
    are expected to not modify it.

    This fixture exists as a speed optimization for tests that only read
    from the model.
    """
    loaded = capellambse.MelodyModel(TEST_ROOT / "5_0" / TEST_MODEL)
    with capellambse.WriteProtector(loaded):
        yield loaded


@pytest.fixture
def model(monkeypatch) -> capellambse.MelodyModel:
    """Return the Capella 5.0 test model."""
    monkeypatch.setattr(sys, "stderr", io.StringIO)
    return capellambse.MelodyModel(TEST_ROOT / "5_0" / TEST_MODEL)


@pytest.fixture
def model_5_2(monkeypatch) -> capellambse.MelodyModel:
    """Return the Capella 5.2 test model."""
    monkeypatch.setattr(sys, "stderr", io.StringIO)
    return capellambse.MelodyModel(TEST_ROOT / "5_2" / TEST_MODEL)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = ["UnsupportedPluginError", "UnsupportedPluginVersionError"]

import collections.abc as cabc
import dataclasses
import re
import typing as t


class UnsupportedPluginError(ValueError):
    """Raised when a plugin is unknown."""


class UnsupportedPluginVersionError(ValueError):
    """Raised when the plugin's version is unsupported."""

    def __str__(self) -> str:
        plugin, ns_plugin = self.args
        assert isinstance(plugin, Plugin) and isinstance(ns_plugin, Plugin)
        return (
            f"Plugin '{plugin.name}' with unsupported version encountered: "
            f"{plugin.version!r} does not match {ns_plugin.version!r}."
        )


@dataclasses.dataclass(frozen=True)
class Plugin:
    """Capella xml-element/plugin with info about name and version."""

    name: str
    version: str | tuple[str, str] | None = None
    viewpoint: str | None = None

    def __post_init__(self) -> None:
        if self.version is not None and self.viewpoint is None:
            raise TypeError("Versioned plugins require a viewpoint")

    @property
    def min_version(self) -> str | None:
        if isinstance(self.version, str):
            return self.version
        if isinstance(self.version, tuple):
            return self.version[0]
        return None

    @property
    def max_version(self) -> str | None:
        if isinstance(self.version, str):
            return self.version
        if isinstance(self.version, tuple):
            return self.version[1]
        return None

    def __le__(self, other: Plugin) -> bool:
        if self.version is None:
            raise AttributeError(f"Plugin '{self.name}' has no version")

        if isinstance(other.version, tuple):
            assert all(isinstance(o, str) for o in other.version)
            vmin, vmax = tuple(map(_tofloat, other.version))
            if isinstance(self.version, tuple):
                vsmin, vsmax = tuple(map(_tofloat, self.version))
                return vsmin <= vmin <= vmax <= vsmax
            else:
                return vmin <= _tofloat(self.version) <= vmax

        assert other.version is not None
        oversion = _tofloat(other.version)
        if isinstance(self.version, tuple):
            vsmin, vsmax = tuple(map(_tofloat, self.version))
            return vsmin <= oversion <= vsmax

        return oversion <= _tofloat(self.version)

    def __str__(self) -> str:
        if suffix := self.min_version:
            return self.name + suffix
        return self.name

    def matches_version(self, value: str | None) -> bool:
        """Check whether a version number is supported by this plugin."""
        if value is None:
            return self.version is None
        if self.version is None:
            return True
        if isinstance(self.version, str):
            return self.version == value

        for mymin, mymax, their in zip(
            self.version[0].split("."),
            self.version[1].split("."),
            value.split("."),
        ):
            if not int(mymin) <= int(their) <= int(mymax):
                return False
        return True


def _tofloat(other: str) -> float:
    """Change 1.x.y...str into 1.x float."""
    version = other.split(".")
    if len(version) > 1:
        version = version[:2]

    return float(".".join(version))


def get_namespace_prefix(url: str) -> str:
    """Map a namespace URL to its symbolic name (the "prefix").

    Parameters
    ----------
    url
        A namespace URL.

    Raises
    ------
    UnsupportedPluginError
        If the given URL does not match any known plugin.
    UnsupportedPluginVersionError
        If the URL matches a known plugin, but the version indicated by
        the URL is outside of the supported range.

    Returns
    -------
    str
        The symbolic name of the namespace matching the URI.
    """
    match = re.match(r"(.*/)((?:\d+\.)*\d+)$", url)
    if match:
        plugin_name, version = match.groups()
    else:
        plugin_name, version = url, None

    matched_plugins = [
        (nskey, nsplugin)
        for nskey, nsplugin in NAMESPACES_PLUGINS.items()
        if nsplugin.name in (plugin_name, url)
    ]
    if not matched_plugins:
        raise UnsupportedPluginError(url)
    if len(matched_plugins) != 1:
        raise RuntimeError(f"Ambiguous namespace {url!r}: {matched_plugins}")
    prefix, plugin = matched_plugins[0]

    if not plugin.matches_version(version):
        raise UnsupportedPluginVersionError(plugin_name, version)

    return prefix


NAMESPACES_PLUGINS: t.Final[cabc.Mapping[str, Plugin]] = {
    "CapellaRequirements": Plugin(
        "http://www.polarsys.org/capella/requirements",
        None,
        "org.polarsys.capella.vp.requirements",
    ),
    "Requirements": Plugin(
        "http://www.polarsys.org/kitalpha/requirements",
        None,
        "org.polarsys.kitalpha.vp.requirements",
    ),
    "concern": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/concern/1.1.0",
    ),
    "description": Plugin("http://www.eclipse.org/sirius/description/1.1.0"),
    "description_1": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/1.1.0"
    ),
    "description_2": Plugin(
        "http://www.eclipse.org/sirius/table/description/1.1.0"
    ),
    "description_3": Plugin(
        "http://www.eclipse.org/sirius/diagram/sequence/description/2.0.0",
    ),
    "diagram": Plugin("http://www.eclipse.org/sirius/diagram/1.1.0"),
    "diagramstyler": Plugin("http://thalesgroup.com/mde/melody/ordering"),
    "ecore": Plugin("http://www.eclipse.org/emf/2002/Ecore"),
    "filter": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/filter/1.1.0",
    ),
    "libraries": Plugin(
        "http://www.polarsys.org/capella/common/libraries/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "metadata": Plugin("http://www.polarsys.org/kitalpha/ad/metadata/1.0.0"),
    "notation": Plugin("http://www.eclipse.org/gmf/runtime/1.0.2/notation"),
    "org.polarsys.capella.core.data.capellacommon": Plugin(
        "http://www.polarsys.org/capella/core/common/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.capellacore": Plugin(
        "http://www.polarsys.org/capella/core/core/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.capellamodeller": Plugin(
        "http://www.polarsys.org/capella/core/modeller/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.cs": Plugin(
        "http://www.polarsys.org/capella/core/cs/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.ctx": Plugin(
        "http://www.polarsys.org/capella/core/ctx/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.epbs": Plugin(
        "http://www.polarsys.org/capella/core/epbs/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.fa": Plugin(
        "http://www.polarsys.org/capella/core/fa/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.information": Plugin(
        "http://www.polarsys.org/capella/core/information/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.information.datatype": Plugin(
        "http://www.polarsys.org/capella/core/information/datatype/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.information.datavalue": Plugin(
        "http://www.polarsys.org/capella/core/information/datavalue/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.interaction": Plugin(
        "http://www.polarsys.org/capella/core/interaction/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.la": Plugin(
        "http://www.polarsys.org/capella/core/la/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.oa": Plugin(
        "http://www.polarsys.org/capella/core/oa/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.pa": Plugin(
        "http://www.polarsys.org/capella/core/pa/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "re": Plugin("http://www.polarsys.org/capella/common/re/1.3.0"),
    "sequence": Plugin("http://www.eclipse.org/sirius/diagram/sequence/2.0.0"),
    "style": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/style/1.1.0"
    ),
    "table": Plugin("http://www.eclipse.org/sirius/table/1.1.0"),
    "viewpoint": Plugin("http://www.eclipse.org/sirius/1.1.0"),
    "xmi": Plugin("http://www.omg.org/XMI"),
    "xsi": Plugin("http://www.w3.org/2001/XMLSchema-instance"),
}
NAMESPACES: t.Final[cabc.Mapping[str, str]] = {
    nskey: str(plugin) for nskey, plugin in NAMESPACES_PLUGINS.items()
}



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=redefined-outer-name
from __future__ import annotations

import base64
import os
import pathlib
import re
import shutil
import subprocess
import sys
from importlib import metadata

import pytest
import requests_mock

import capellambse
from capellambse.filehandler import gitlab_artifacts

# pylint: disable-next=relative-beyond-top-level, useless-suppression
from .conftest import TEST_MODEL, TEST_ROOT  # type: ignore[import-untyped]

TEST_MODEL_5_0 = TEST_ROOT / "5_0" / TEST_MODEL

DUMMY_SVG = b'<svg xmlns="http://www.w3.org/2000/svg"/>'
DUMMY_PNG = base64.standard_b64decode(
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAACklEQVR4nGMAAQAABQABDQott"
    "AAAAABJRU5ErkJggg=="
)


@pytest.fixture(autouse=True, scope="function")
def glart_clear_env(monkeypatch):
    for i in list(os.environ):
        if i.startswith("CI_"):
            monkeypatch.delenv(i)


@pytest.mark.parametrize(
    "path",
    [
        pytest.param(str(TEST_MODEL_5_0), id="From string"),
        pytest.param(TEST_MODEL_5_0, id="From path"),
    ],
)
def test_model_loading_via_LocalFileHandler(path: str | pathlib.Path):
    capellambse.MelodyModel(path)


@pytest.mark.parametrize("suffix", [".afm", ".capella"])
def test_model_loading_with_invalid_entrypoint_fails(suffix: str):
    with pytest.raises(ValueError, match="(?i)invalid entrypoint"):
        capellambse.MelodyModel(TEST_MODEL_5_0.with_suffix(suffix))


def test_model_loading_via_GitFileHandler():
    path = "git+" + pathlib.Path.cwd().as_uri()
    capellambse.MelodyModel(
        path, entrypoint="tests/data/melodymodel/5_0/Melody Model Test.aird"
    )
    assert not pathlib.Path.cwd().joinpath("capellambse.lock").exists()


def test_model_loading_via_GitFileHandler_invalid_uri():
    path = "git+not-a-valid-uri"
    with pytest.raises(subprocess.CalledProcessError):
        capellambse.MelodyModel(
            path,
            entrypoint="tests/data/melodymodel/5_0/Melody Model Test.aird",
        )
    assert path


def test_model_loading_from_badpath_raises_FileNotFoundError():
    badpath = TEST_ROOT / "Missing.aird"
    with pytest.raises(FileNotFoundError):
        capellambse.MelodyModel(badpath)


def test_split_protocol_returns_path_objects_unchanged():
    expected = pathlib.Path.cwd()

    handler, actual = capellambse.filehandler.split_protocol(expected)

    assert handler == "file"
    assert actual == expected


@pytest.mark.parametrize(
    ["url", "expected"],
    (
        [
            ("/path/to/file", pathlib.Path("/path/to/file")),
            ("file:///path/to/file", pathlib.Path("/path/to/file")),
            ("file://localhost/path/to/file", pathlib.Path("/path/to/file")),
        ]
        if not sys.platform.startswith("win")
        else [
            ("C:/path/to/file", pathlib.Path("C:/path/to/file")),
            (r"C:\path\to\file", pathlib.Path(r"C:\path\to\file")),
            ("file:///C:/path/to/file", pathlib.Path("C:/path/to/file")),
            (
                "file://localhost/C:/path/to/file",
                pathlib.Path("C:/path/to/file"),
            ),
        ]
    ),
)
def test_split_protocol_converts_file_urls_to_paths(url, expected):
    handler, actual = capellambse.filehandler.split_protocol(url)

    assert handler == "file"
    assert actual == expected


@pytest.mark.parametrize(
    "url",
    [
        "myhost:myrepo.git",
        "git@host:",
        "git@host:repo.git",
        "git@host:path/to/repo",
        "git@host:/path/to/repo",
    ],
)
def test_split_protocol_recognizes_scp_style_uris_as_git(url: str):
    handler, actual = capellambse.filehandler.split_protocol(url)

    assert handler == "git"
    assert actual == url


@pytest.mark.parametrize(
    "url",
    [
        pytest.param("http://domain.invalid/path", id="http"),
        pytest.param("https://domain.invalid/path", id="https"),
        pytest.param("ftp://domain.invalid/path", id="ftp"),
        pytest.param("ftps://domain.invalid/path", id="ftps"),
        pytest.param("sftp://domain.invalid/path", id="sftp"),
        pytest.param("ssh://domain.invalid/path", id="ssh"),
    ],
)
def test_split_protocol_does_not_change_simple_urls(url: str):
    handler, actual = capellambse.filehandler.split_protocol(url)

    assert handler == url.split(":", 1)[0]
    assert actual == url


class FakeEntrypoint:
    def __init__(self, expected_name: str, expected_url: str):
        self._expected_name = expected_name
        self._expected_url = expected_url

    @property
    def name(self):
        class AlwaysEqual:
            def __eq__(_, name):  # pylint: disable=E,I
                nonlocal self
                assert name == self._expected_name
                return True

        return AlwaysEqual()

    def load(self):
        def filehandler(url: str | pathlib.Path):
            if isinstance(url, pathlib.Path):
                assert url == pathlib.Path(self._expected_url)
            else:
                assert str(url) == self._expected_url

        return filehandler

    @classmethod
    def patch(cls, monkeypatch, expected_name, expected_url):
        def entry_points(group, name):
            assert group == "capellambse.filehandler"
            assert name == expected_name
            return (cls(expected_name, expected_url),)

        monkeypatch.setattr(metadata, "entry_points", entry_points)


def test_a_single_protocol_is_not_swallowed_by_get_filehandler(
    monkeypatch,
):
    FakeEntrypoint.patch(monkeypatch, "testproto", "testproto://url")

    capellambse.get_filehandler("testproto://url")


def test_a_wrapping_protocol_separated_by_plus_is_stripped(monkeypatch):
    FakeEntrypoint.patch(monkeypatch, "realproto", "wrappedproto://url")

    capellambse.get_filehandler("realproto+wrappedproto://url")


@pytest.mark.parametrize(
    "path",
    (
        [
            "/data/model/model.aird",
        ]
        if not sys.platform.startswith("win")
        else [
            r"S:\model\model.aird",
            r"S:/model/model.aird",
            r"\\?\S:\model\model.aird",
            r"\\modelserver\model\model.aird",
        ]
    ),
)
def test_plain_file_paths_are_recognized_as_file_protocol(monkeypatch, path):
    FakeEntrypoint.patch(monkeypatch, "file", path)

    capellambse.get_filehandler(path)


@pytest.mark.parametrize(
    "url",
    [
        "myhost:myrepo.git",
        "git@host:",
        "git@host:repo.git",
        "git@host:path/to/repo",
        "git@host:/path/to/repo",
    ],
)
def test_the_scp_short_form_is_recognized_as_git_protocol(monkeypatch, url):
    FakeEntrypoint.patch(monkeypatch, "git", url)

    capellambse.get_filehandler(url)


@pytest.mark.parametrize(
    "link",
    [
        "xtype MelodyModelTest.aird#00000000-0000-4000-0000-000000000000",
        "MelodyModelTest.aird#00000000-0000-4000-0000-000000000000",
        "#00000000-0000-4000-0000-000000000000",
        "xtype MelodyModel%20Test.aird#00000000-0000-4000-0000-000000000000",
        "MelodyModel%20Test.aird#00000000-0000-4000-0000-000000000000",
    ],
)
def test_MelodyLoader_follow_link_finds_target(link: str):
    loader = capellambse.loader.MelodyLoader(TEST_MODEL_5_0)

    with pytest.raises(KeyError):
        loader.follow_link(None, link)


@pytest.mark.parametrize(
    ["path", "subdir", "req_url"],
    [
        (
            "https://example.com/~user",
            "/",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/~user/%s",
            "/",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/?file=%q",
            "/",
            "https://example.com/?file=demo%2Fmy%20model.aird",
        ),
        (
            "https://example.com/",
            "~user",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/%s",
            "~user",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/?file=%q",
            "~user",
            "https://example.com/?file=~user%2Fdemo%2Fmy%20model.aird",
        ),
        (
            "https://example.com/?dir=%d&file=%n&type=%e",
            "/",
            "https://example.com/?dir=demo&file=my%20model&type=aird",
        ),
    ],
)
def test_http_file_handler_replaces_percent_escapes(
    requests_mock: requests_mock.Mocker, path: str, subdir: str, req_url: str
) -> None:
    endpoint = requests_mock.get(req_url)

    file_handler = capellambse.get_filehandler(path, subdir=subdir)
    file_handler.open("demo/my model.aird", "rb").close()

    assert endpoint.called_once


def test_http_file_handler_hands_auth_to_server(
    requests_mock: requests_mock.Mocker,
) -> None:
    username = "testuser"
    password = "testpassword"
    auth_header = base64.standard_b64encode(
        f"{username}:{password}".encode()
    ).decode("ascii")
    expected_headers = {"Authorization": f"Basic {auth_header}"}
    endpoint = requests_mock.get(
        "https://example.com/test.svg",
        request_headers=expected_headers,
    )

    file_handler = capellambse.get_filehandler(
        "https://example.com", username=username, password=password
    )
    file_handler.open("test.svg", "rb").close()

    assert endpoint.called_once


def test_http_file_handlers_passed_through_custom_headers(
    requests_mock: requests_mock.Mocker,
) -> None:
    expected_headers = {"X-Test-Header": "PASSED"}
    endpoint = requests_mock.get(
        "https://example.com/test.svg",
        request_headers=expected_headers,
    )

    file_handler = capellambse.get_filehandler(
        "https://example.com", headers=expected_headers
    )
    file_handler.open("test.svg", "rb").close()

    assert endpoint.called_once


def test_gitlab_artifacts_handler_uses_public_gitlab_when_no_hostname_given(
    requests_mock: requests_mock.Mocker,  # pylint: disable=unused-argument
) -> None:
    hdl = capellambse.get_filehandler(
        "glart://",
        project=1,
        branch="some-branch",
        token="my-access-token",
        job=3,
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


def test_gitlab_artifacts_handler_resolves_project_to_numeric_id(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/homegroup%2fdemo-project",
        json={"id": 1},
    )

    hdl = capellambse.get_filehandler(
        "glart://",
        project="homegroup/demo-project",
        branch="some-branch",
        token="my-access-token",
        job=3,
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


def test_gitlab_artifacts_handler_looks_up_job_id_given_a_job_name(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs",
        json=[
            {
                "id": 3,
                "name": "make-stuff",
                "pipeline": {"ref": "somebranch"},
                "artifacts": [{"file_type": "archive"}],
            },
        ],
    )

    hdl = capellambse.get_filehandler(
        "glart://",
        project=1,
        branch="somebranch",
        token="my-access-token",
        job="make-stuff",
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


def test_gitlab_artifacts_handler_errors_if_no_job_with_artifacts_was_found(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs",
        json=[],
    )

    with pytest.raises(RuntimeError, match="recent successful"):
        capellambse.get_filehandler(
            "glart://",
            project=1,
            branch="somebranch",
            token="my-access-token",
            job="make-stuff",
        )


def test_gitlab_artifacts_handler_handles_pagination_when_searching_for_jobs(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs",
        headers={
            "Link": (
                "<https://gitlab.com/api/v4/projects/1/jobs/page/2>;"
                ' rel="next"'
            )
        },
        json=[
            {
                "id": 2,
                "name": "make-stuff",
                "pipeline": {"ref": "otherbranch"},
                "artifacts": [],
            },
        ],
    )
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs/page/2",
        json=[
            {
                "id": 3,
                "name": "make-stuff",
                "pipeline": {"ref": "somebranch"},
                "artifacts": [{"file_type": "archive"}],
            },
        ],
    )

    hdl = capellambse.get_filehandler(
        "glart://",
        project=1,
        branch="somebranch",
        token="my-access-token",
        job="make-stuff",
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


@pytest.fixture
def model_path_with_patched_version(
    request: pytest.FixtureRequest, tmp_path: pathlib.Path
) -> pathlib.Path:
    """Indirect parametrized fixture for version patched model.

    Parameters
    ----------
    request
        Special pytest fixture for access to test context, exposing the
        ``param`` attribute on indirect parametrization.
    tmp_path
        Pytest fixture giving a temporary path.

    Returns
    -------
    aird_path
        Path to Capella ``.aird`` file.
    """
    tmp_dest = tmp_path / "model"
    ignored = shutil.ignore_patterns("*.license")
    shutil.copytree(TEST_MODEL_5_0.parent, tmp_dest, ignore=ignored)
    request_param: str | tuple[str, str] = request.param
    for suffix in (".aird", ".capella"):
        model_file = (tmp_dest / TEST_MODEL).with_suffix(suffix)
        if isinstance(request_param, tuple):
            re_params = request_param
        else:
            re_params = (r"5\.0\.0", request_param)

        patched = re.sub(*re_params, model_file.read_text(encoding="utf-8"))
        model_file.write_text(patched, encoding="utf-8")
    return model_file.with_suffix(".aird")


@pytest.mark.parametrize(
    "model_path_with_patched_version",
    ["1.3.0", "1.4.2", "6.1.0"],
    indirect=True,
)
def test_loading_model_with_unsupported_version_fails(
    model_path_with_patched_version: pathlib.Path,
) -> None:
    with pytest.raises(capellambse.UnsupportedPluginVersionError):
        capellambse.MelodyModel(model_path_with_patched_version)


@pytest.mark.parametrize(
    "model_path_with_patched_version",
    [("http://www.eclipse.org/sirius/1.1.0", "Unknown")],
    indirect=True,
)
def test_loading_model_with_unsupported_plugin_fails(
    model_path_with_patched_version: pathlib.Path,
) -> None:
    with pytest.raises(capellambse.UnsupportedPluginError):
        capellambse.MelodyModel(model_path_with_patched_version)


def test_model_info_contains_viewpoints_and_capella_version() -> None:
    loader = capellambse.loader.MelodyLoader(TEST_MODEL_5_0)

    info = loader.get_model_info()

    assert hasattr(info, "viewpoints")
    assert info.viewpoints["org.polarsys.capella.core.viewpoint"] == "5.0.0"
    assert info.capella_version == "5.0.0"


@pytest.mark.parametrize(
    ["format", "content"],
    [
        pytest.param("svg", DUMMY_SVG, id="svg"),
        pytest.param("png", DUMMY_PNG, id="png"),
    ],
)
def test_model_loads_diagrams_from_cache_by_uuid(
    tmp_path: pathlib.Path, format: str, content: bytes
):
    model = capellambse.MelodyModel(TEST_MODEL_5_0, diagram_cache=tmp_path)
    dg = model.diagrams[0]
    tmp_path.joinpath(f"{dg.uuid}.{format}").write_bytes(content)

    rendered = dg.render(format)
    if isinstance(rendered, str):
        rendered = rendered.encode("utf-8")

    assert rendered == content


def test_model_will_refuse_to_render_diagrams_if_diagram_cache_was_given(
    tmp_path: pathlib.Path,
):
    model = capellambse.MelodyModel(TEST_MODEL_5_0, diagram_cache=tmp_path)

    with pytest.raises(RuntimeError, match="not in cache"):
        model.diagrams[0].render("svg")


def test_model_will_fall_back_to_rendering_internally_despite_the_cache_if_told_to(
    tmp_path: pathlib.Path,
):
    model = capellambse.MelodyModel(
        TEST_MODEL_5_0,
        diagram_cache=tmp_path,
        fallback_render_aird=True,
    )

    assert model.diagrams[0].render("svg")


def test_model_diagram_visible_nodes_can_be_accessed_when_a_cache_was_specified(
    tmp_path: pathlib.Path,
):
    model = capellambse.MelodyModel(TEST_MODEL_5_0, diagram_cache=tmp_path)

    assert model.diagrams[0].nodes



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=redefined-outer-name
import pathlib
import shutil
import tempfile

import pytest

from capellambse import loader, pvmt

TEST_ROOT = pathlib.Path(__file__).parent / "data" / "pvmt"
MODEL_FILE = "PVMTTest.aird"


@pytest.fixture
def model():
    global TEST_ROOT

    orig_test_root = TEST_ROOT
    with tempfile.TemporaryDirectory() as tempdir:
        TEST_ROOT = pathlib.Path(tempdir, "model")
        shutil.copytree(orig_test_root, TEST_ROOT)

        yield loader.MelodyLoader(TEST_ROOT / MODEL_FILE)

    TEST_ROOT = orig_test_root


@pytest.fixture
def pvext(model):
    return pvmt.load_pvmt_from_model(model)


class TestPVMTBase:
    """Tests for basic PVMT functionality."""

    domain_uuid = "02e0c435-f085-471f-9f6e-e12fe5f27687"
    domain_name = "Computer"

    enum_names = {
        "03e5e5ae-0a61-473d-a792-e003ce601ff4": "Cable Type",
        "ae279311-6a75-40b7-9bf3-a19c136d0573": "ComponentType",
    }
    enum_values = {
        "03e5e5ae-0a61-473d-a792-e003ce601ff4": ["UNSET", "Trace", "Cable"],
        "ae279311-6a75-40b7-9bf3-a19c136d0573": [
            "UNSET",
            "Core Component",
            "External",
        ],
    }

    def test_domains(self, pvext):
        assert {k: v.name for k, v in pvext.items()} == {
            self.domain_uuid: self.domain_name,
            "12a02f1b-8b97-4188-95ea-2afd377c4c41": "Out of scope",
            "ae15f861-2f15-40d2-a1b9-57dfdf7fabae": "In scope",
            "c2d3e3a3-f9bb-4ff7-964d-2109c7065bc2": "External Data",
        }

    def test_enums(self, pvext):
        model_enums = pvext[self.domain_uuid].enums
        assert {k: v.name for k, v in model_enums.items()} == self.enum_names

        for enum_uuid, vals in self.enum_values.items():
            actual_vals = {
                v["name"] for _, v in model_enums[enum_uuid].items()
            }
            assert actual_vals == set(vals)

    def test_enum_defaults(self, pvext):
        group = "0fb1fbfb-43a6-4f96-9ec3-2cb83d80c702"
        pv_id = "f3f9b7ec-f00d-4b2f-8a48-900a22c43853"
        epv = pvext[self.domain_uuid][group][pv_id]
        assert epv.default_value["name"] == "UNSET"

    def test_groups(self, pvext):
        model_groups = pvext[self.domain_uuid].groups
        assert {g.name for g in model_groups} == {
            "Components",
            "Cables",
            "Physical Cables",
        }

    def test_int(self, pvext):
        group = "cdc24b8b-3c4c-4513-9aad-99c0f7d884dc"
        pv_id = "7130121b-0242-49b7-885c-44b868c4d7b7"
        ipv = pvext[self.domain_uuid][group][pv_id]

        assert ipv.name == "Price"
        assert ipv.unit == "€"

    def test_apply_outofscope(self, model, pvext):
        elem = model["d32caffc-b9a1-448e-8e96-65a36ba06292"]
        domain = pvext["12a02f1b-8b97-4188-95ea-2afd377c4c41"]
        for group in domain.groups:
            with pytest.raises(pvmt.ScopeError):
                pvext.get_element_pv(
                    elem, f"{domain.name}.{group.name}", create=True
                )

    def test_apply_inscope(self, model, pvext):
        elem = model["d32caffc-b9a1-448e-8e96-65a36ba06292"]
        domain = pvext["ae15f861-2f15-40d2-a1b9-57dfdf7fabae"]
        for group in domain.groups:
            pvext.get_element_pv(
                elem, f"{domain.name}.{group.name}", create=True
            )


class TestAppliedPropertyValueGroup:
    """Tests for all methods of the ``AppliedPropertyValueGroup`` object."""

    elem_uuid = "d32caffc-b9a1-448e-8e96-65a36ba06292"
    pvg_name = "Computer.Physical Cables"

    def test___iter__(self, model, pvext):
        elem_pv = pvext.get_element_pv(model[self.elem_uuid], self.pvg_name)
        keys = set(elem_pv)
        assert keys == {"In Stock", "Label", "Plugs per side", "Price"}

    def test___len__(self, model, pvext):
        elem_pv = pvext.get_element_pv(model[self.elem_uuid], self.pvg_name)
        assert len(elem_pv) == 4

    def test___getitem__(self, model, pvext):
        elem_pv = pvext.get_element_pv(model[self.elem_uuid], self.pvg_name)
        assert elem_pv["In Stock"] is True
        assert elem_pv["Label"] == "DisplayPort_1"
        assert elem_pv["Plugs per side"] == 1
        assert elem_pv["Price"] == 14.99
        with pytest.raises(KeyError):
            elem_pv["price"]  # pylint: disable=pointless-statement

    def test___setitem___modify(self, model, pvext):
        elem = model[self.elem_uuid]
        elem_pv = pvext.get_element_pv(elem, self.pvg_name)
        pv_cable = pvext.get_element_pv(elem, "Computer.Cables")
        pv_cable["CableType"] = "Trace"
        elem_pv["Price"] = 10
        elem_pv["In Stock"] = False

        assert pv_cable["CableType"] == "Trace"
        assert elem_pv["Price"] == 10.0
        assert elem_pv["In Stock"] is False

        with pytest.raises(KeyError):
            elem_pv["price"] = 10.0

        with pytest.raises(ValueError):
            elem_pv["Price"] = "Not a float"

    def test___delitem_____setitem__(self, model, pvext):
        elem_pv = pvext.get_element_pv(model[self.elem_uuid], self.pvg_name)
        del elem_pv["Price"]
        assert "Price" not in elem_pv

        with pytest.raises(KeyError):
            del elem_pv["Price"]
        with pytest.raises(KeyError):
            del elem_pv["nonexistent_key"]
        with pytest.raises(KeyError):
            elem_pv["price"] = 10.0

        elem_pv["Price"] = 25.0
        assert "Price" in elem_pv
        assert elem_pv["Price"] == 25.0

    def test___contains__(self, model, pvext):
        elem_pv = pvext.get_element_pv(model[self.elem_uuid], self.pvg_name)
        assert "Price" in elem_pv
        assert "In Stock" in elem_pv
        assert "price" not in elem_pv
        assert "InStock" not in elem_pv

    def test_copy(self, model, pvext):
        elem_pv = pvext.get_element_pv(model[self.elem_uuid], self.pvg_name)
        copy = elem_pv.copy()
        expected = {
            "In Stock": True,
            "Label": "DisplayPort_1",
            "Plugs per side": 1,
            "Price": 14.99,
        }
        assert expected == copy
        assert elem_pv is not copy

    def test_get(self, model, pvext):
        elem_pv = pvext.get_element_pv(model[self.elem_uuid], self.pvg_name)
        assert elem_pv.get("Price") == 14.99
        assert elem_pv.get("In Stock") is True

        sentinel = object()
        assert elem_pv.get("price", sentinel) is sentinel


class TestAppliedPropertyValueGroupXML:
    """Tests that rely on writing back and comparing the output XML."""

    expect_root = TEST_ROOT / "expected-output"

    elem_uuid = "d32caffc-b9a1-448e-8e96-65a36ba06292"

    def compare_xml(self, model, expected_file):
        pvmt_model_path = (TEST_ROOT / MODEL_FILE).with_suffix(".capella")
        expected_model_path = self.expect_root / expected_file

        model.save()
        actual = pvmt_model_path.read_text(encoding="utf-8")
        expected = expected_model_path.read_text(encoding="utf-8")

        assert actual == expected

    def test_apply(self, model, pvext, monkeypatch):
        call_count = 0

        def mock_generate_uuid(*__, **_):
            nonlocal call_count
            call_count += 1
            return f"00000000-0000-0000-0000-{call_count:012x}"

        monkeypatch.setattr(
            "capellambse.loader.MelodyLoader.generate_uuid",
            mock_generate_uuid,
        )

        elem = model[self.elem_uuid]

        obj_ids = pvext.get_element_pv(
            elem, "External Data.Object IDs", create=True
        )
        # By this point, the group and its child must both already
        # exist, each with their own UUID.
        assert call_count == 2

        obj_ids["Object ID"] = "CABLE-0001"
        self.compare_xml(model, "apply.capella")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections.abc as cabc
import errno
import io
import logging
import os
import pathlib
import typing as t
import weakref
import zipfile

from capellambse import helpers
from capellambse.loader import modelinfo

from . import abc, get_filehandler

LOGGER = logging.getLogger(__name__)


class ZipFileHandler(abc.FileHandler):
    """File handler that can read from zip files.

    Parameters
    ----------
    path
        Location of the zip file. May contain a nested path, like
        ``zip+https://host.name/%s``, which will be resolved using an
        appropriate file handler.

        If ``zipname`` is not passed or is None, the path may include the zip
        file name as well, using either ``!`` or ``/`` as separator. For
        example, the following calls are equivalent:

        .. code-block:: python

           ZipFileHandler("zip+https://host.name/path/to/file.zip")
           ZipFileHandler("zip+https://host.name/path/to!file.zip")
           ZipFileHandler("zip+https://host.name/path/to/%s!file.zip")

        .. note::

           The ``%s`` replacement shown in this example is done by the
           underlying :class:`~capellambse.filehandler.http.HTTPFileHandler`,
           which is used to fetch the nested ``https://`` URL. Other nested
           protocols may require different syntax.)

        .. note::

           It is currently not possible to pass down arbitrary arguments to the
           underlying FileHandler other than the ``path``.
    zipname
        Name of the zip file in the above path.
    subdir
        A subdirectory inside the zip file to use as base directory.

        If the zip file contains only a single directory entry and no
        other files at the root, this directory is used as default. Pass
        ``subdir="."`` explicitly to override this behaviour.
    """

    def __init__(
        self,
        path: str | os.PathLike,
        zipname: str | pathlib.PurePosixPath | None = None,
        subdir: str | pathlib.PurePosixPath | None = None,
    ) -> None:
        super().__init__(path, subdir=subdir or ".")
        path = os.fspath(path)
        if path.startswith("zip://"):
            path = path.replace("zip", "file", 1)
        if zipname is None:
            if "!" in path:
                path, zipname = path.split("!", 1)
            elif "/" in path:
                path, zipname = path.rsplit("/", 1)
            else:
                path, zipname = ".", path

        slave = get_filehandler(path)
        file = slave.read_file(zipname)
        # pylint: disable-next=consider-using-with
        self.__file = zipfile.ZipFile(io.BytesIO(file))
        # pylint: disable-next=unused-private-member
        self.__fnz = weakref.finalize(self, self.__file.close)

        if subdir is None:
            subdir = os.path.commonprefix(self.__file.namelist())
            subdir = subdir[: subdir.rfind("/") + 1].rstrip("/")
            LOGGER.debug("Auto-selected base directory: %s", subdir)
            self.subdir = pathlib.PurePosixPath(subdir)

    def get_model_info(self) -> modelinfo.ModelInfo:
        return modelinfo.ModelInfo(url=os.fspath(self.path))

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.IO[bytes]:
        if "w" in mode:
            raise ValueError("Writing to zip files is not supported")

        filename = helpers.normalize_pure_path(filename, base=self.subdir)
        try:
            return self.__file.open(str(filename), "r")
        except KeyError:
            raise FileNotFoundError(errno.ENOENT, filename) from None

    @property
    def rootdir(self) -> abc.FilePath[ZipFileHandler]:
        return abc.FilePath(self, pathlib.PurePosixPath("."))

    def iterdir(
        self, path: str | pathlib.PurePosixPath = ".", /
    ) -> cabc.Iterator[abc.FilePath[ZipFileHandler]]:
        path, zpath = _normalize_path(path, self.subdir)
        for zsub in zipfile.Path(self.__file, zpath).iterdir():
            yield abc.FilePath(self, path / zsub.name)

    def is_dir(self, path: str | pathlib.PurePosixPath, /) -> bool:
        _, zpath = _normalize_path(path, self.subdir)
        return zipfile.Path(self.__file, zpath).is_dir()

    def is_file(self, path: str | pathlib.PurePosixPath, /) -> bool:
        _, zpath = _normalize_path(path, self.subdir)
        return zipfile.Path(self.__file, zpath).is_file()


def _normalize_path(
    path: str | pathlib.PurePosixPath, base: pathlib.PurePosixPath, /
) -> tuple[pathlib.PurePosixPath, str]:
    path = helpers.normalize_pure_path(path)
    zpath = str(base / path) + "/"
    if zpath == "./":
        zpath = ""
    return path, zpath



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Parser entry point for visual elements.

Visual elements, contrary to semantic ones, do not exist in the melody
files at all. They are used for things such as Notes and inter-diagram
hyperlinks.
"""
from __future__ import annotations

import collections.abc as cabc
import logging

from capellambse import diagram

from . import _common as c
from . import _edge_factories, _styling

LOGGER = logging.getLogger(__name__)


def from_xml(ebd: c.ElementBuilder) -> diagram.DiagramElement:
    """Deserialize a visual element."""
    el_type = ebd.data_element.attrib[c.ATT_XMT].split(":")[-1]
    if el_type not in VISUAL_TYPES:
        LOGGER.error("Unknown visual element type, skipping: %r", el_type)
        raise c.SkipObject()
    factory = VISUAL_TYPES[el_type]
    if factory is None:
        raise c.SkipObject()
    return factory(ebd)


def connector_factory(ebd: c.ElementBuilder) -> diagram.Edge:
    """Create a Connector.

    A Connector is an edge that connects Notes to arbitrary elements.
    """
    styleclass = "Connector"
    bendpoints, source, target = _edge_factories.extract_bendpoints(ebd)
    styleoverrides = _styling.apply_style_overrides(
        ebd.target_diagram.styleclass, styleclass, ebd.data_element
    )
    edge = diagram.Edge(
        bendpoints,
        uuid=ebd.data_element.attrib[c.ATT_XMID],
        styleclass=styleclass,
        # <https://github.com/python/mypy/issues/8136#issuecomment-565387901>
        styleoverrides=styleoverrides,  # type: ignore[arg-type]
    )
    assert edge.uuid is not None

    if isinstance(target, diagram.Box):
        _edge_factories.snaptarget(edge.points, -1, -2, target)
        target.add_context(edge.uuid)
    if isinstance(source, diagram.Box):
        _edge_factories.snaptarget(edge.points, 0, 1, source)
        source.add_context(edge.uuid)

    return edge


def shape_factory(ebd: c.ElementBuilder) -> diagram.Box:
    """Create a Shape."""
    assert ebd.target_diagram.styleclass is not None

    uid = ebd.data_element.attrib[c.ATT_XMID]
    element = ebd.data_element.get("element")
    if element is not None:
        label = ebd.melodyloader[element].attrib["name"]
        description = ebd.data_element.get("description", "")
    else:
        label = ebd.data_element.get("description", "")
        description = None
    parent_elm = ebd.data_element.getparent()
    while parent_elm is not None and parent_elm.tag == "children":
        parent_uid = (
            parent_elm.attrib.get("element")
            or parent_elm.attrib.get(c.ATT_XMID)
            or ""
        )
        try:
            parent = ebd.target_diagram[parent_uid]
        except KeyError:
            parent_elm = parent_elm.getparent()
        else:
            assert isinstance(parent, diagram.Box)
            refpos = parent.pos
            break
    else:
        parent = None
        refpos = diagram.Vector2D(0, 0)

    try:
        layout = next(ebd.data_element.iterchildren("layoutConstraint"))
    except StopIteration:
        raise ValueError(
            "No layoutConstraint found for element {uid}"
        ) from None

    pos = refpos + (
        int(layout.attrib.get("x", "0")),
        int(layout.attrib.get("y", "0")),
    )
    size = diagram.Vector2D(
        int(layout.attrib.get("width", "0")),
        int(layout.attrib.get("height", "0")),
    )

    if element is not None:
        styleclass = "RepresentationLink"
    else:
        styleclass = "Note"

    styleoverrides = _styling.apply_visualelement_styles(
        ebd.target_diagram.styleclass, f"Box.{styleclass}", ebd.data_element
    )

    return diagram.Box(
        pos,
        size,
        label=label,
        description=description,
        uuid=uid,
        parent=parent,
        styleclass=styleclass,
        # <https://github.com/python/mypy/issues/8136#issuecomment-565387901>
        styleoverrides=styleoverrides,  # type: ignore[arg-type]
        minsize=diagram.Vector2D(100, 54),
    )


VISUAL_TYPES: dict[
    str, cabc.Callable[[c.ElementBuilder], diagram.DiagramElement]
] = {
    "BasicDecorationNode": c.SkipObject.raise_,
    "Connector": connector_factory,
    # Nodes are actually semantic elements. If one got through to here,
    # it's an internal node, not an element root.
    "Node": c.SkipObject.raise_,
    "Shape": shape_factory,
}



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Factory functions for different types of Boxes."""
from __future__ import annotations

import collections
import functools
import itertools
import re
import typing as t

import markupsafe

from capellambse import diagram, helpers

from . import _common as C
from . import _filters, _styling

_T = t.TypeVar("_T", bound=diagram.Box)


@t.overload
def generic_factory(
    seb: C.SemanticElementBuilder, *, minsize: diagram.Vector2D = ...
) -> diagram.Box: ...


@t.overload
def generic_factory(
    seb: C.SemanticElementBuilder,
    *,
    boxtype: type[_T] | functools.partial[_T],
    minsize: diagram.Vector2D = ...,
) -> _T: ...


def generic_factory(
    seb: C.SemanticElementBuilder,
    *,
    boxtype: (
        type[diagram.Box] | type[_T] | functools.partial[_T]
    ) = diagram.Box,
    minsize: diagram.Vector2D = diagram.Vector2D(148, 69),
) -> _T:
    """Construct a Box from the diagram XML."""
    diag_parent = seb.diag_element.getparent()
    assert diag_parent is not None
    if diag_parent is not seb.diagram_tree:
        parent_uid = diag_parent.get("uid")
    else:
        parent_uid = None

    if parent_uid is None:
        parent = None
        refpos = diagram.Vector2D(0, 0)
    else:
        try:
            # reference position (top left of parent box)
            parent = seb.target_diagram[parent_uid]
            refpos = parent.bounds.pos
        except KeyError:
            C.LOGGER.error(
                "Parent not in diagram, cannot draw box with uid %r",
                seb.data_element.attrib["element"],
            )
            raise C.SkipObject() from None
    assert parent is None or isinstance(parent, diagram.Box)

    try:
        layout = next(seb.data_element.iterchildren("layoutConstraint")).attrib
        ostyle = next(seb.diag_element.iterchildren("ownedStyle"))
    except StopIteration:
        C.LOGGER.error(
            "Cannot find style or layout for %r",
            seb.data_element.attrib["element"],
        )
        raise C.SkipObject() from None

    box_is_port = seb.diag_element.tag == "ownedBorderedNodes"
    box_is_symbol = ostyle.get("workspacePath") is not None

    pos = refpos + (int(layout.get("x", 0)), int(layout.get("y", 0)))
    if box_is_port:
        size = C.PORT_SIZE
    else:
        size = diagram.Vector2D(
            int(layout.get("width", 0)), int(layout.get("height", 0))
        )
        style_type = ostyle.attrib[C.ATT_XMT].split(":")[-1]
        if style_type == "FlatContainerStyle":
            # Remove drop shadows
            size -= (2 if size.x >= 2 else 0, 2 if size.y >= 2 else 0)
    styleoverrides = _styling.apply_style_overrides(
        seb.target_diagram.styleclass, f"Box.{seb.styleclass}", ostyle
    )

    label = seb.melodyobjs[0].attrib.get("name", "")
    pos += (
        int(seb.diag_element.attrib.get("width", (5, -1)[box_is_port])),
        int(seb.diag_element.attrib.get("height", (5, -1)[box_is_port])),
    )

    if box_is_port and parent is not None:
        parent.add_context(seb.data_element.attrib["element"])

    floating_labels: list[diagram.Box] = []
    if label:
        if box_is_port:
            floating_labels.append(_make_portlabel(pos, size, label, seb))
        elif box_is_symbol:
            floating_labels.append(
                _make_free_floating_label(pos, size, label, seb)
            )

    if floating_labels:
        label = ""

    box = boxtype(
        pos,
        size,
        label=label,
        floating_labels=floating_labels,
        collapsed=_is_collapsed(seb),
        port=box_is_port,
        uuid=seb.data_element.attrib["element"],
        styleclass=seb.styleclass,
        # <https://github.com/python/mypy/issues/8136#issuecomment-565387901>
        styleoverrides=styleoverrides,  # type: ignore[arg-type]
        minsize=minsize,
    )
    if box_is_symbol:
        box.JSON_TYPE = "symbol"
        box.minsize = (30, 30)
    _filters.setfilters(seb, box)
    box.parent = parent
    return t.cast(_T, box)


def generic_stacked_factory(seb: C.SemanticElementBuilder) -> C.StackingBox:
    """Construct a Box whose children stack."""
    child_layout = (
        seb.diag_element.get("childrenPresentation") or "VerticalStack"
    )
    return generic_factory(
        seb,
        boxtype=functools.partial(
            C.StackingBox, features=[], stacking_mode=child_layout
        ),
    )


def class_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Create a Class.

    Classes contain multiple `Property` sub-elements. These aren't
    actual boxes, but should instead be shown as part of the Class'
    regular label text.
    """
    box = generic_factory(seb, minsize=diagram.Vector2D(93, 43))

    if seb.melodyobjs[0].attrib.get("abstract", "false") == "true":
        box.styleoverrides["text_font-style"] = "italic"

    features = collections.defaultdict[str, list[str]](list)
    for feature in seb.melodyobjs[0].iterchildren("ownedFeatures"):
        feat_name = feature.attrib["name"]
        feat_type = feature.attrib[C.ATT_XST].split(":")[-1]

        abstract_type_link = feature.get("abstractType")
        if abstract_type_link is not None:
            abstract_type = seb.melodyloader.follow_link(
                seb.melodyobjs[0], abstract_type_link
            )
        else:
            abstract_type = None

        if feat_type == "Property":
            if feature.attrib.get("aggregationKind") is not None:
                continue

            if feature.attrib.get("isDerived") is not None:
                feat_name = f"/{feat_name}"

            if abstract_type is not None:
                feat_name += f" : {abstract_type.attrib['name']}"
        elif feat_type == "Service":
            params = []
            ret = []
            throw = []

            for param in feature.iterchildren("ownedParameters"):
                param_type = param.attrib[C.ATT_XST].split(":")[-1]
                if param_type != "Parameter":
                    C.LOGGER.warning(
                        "Unknown parameter type %r for service %r",
                        param_type,
                        feat_name,
                    )
                    continue
                param_name = param.attrib["name"]
                param_abstrtype_link = param.get("abstractType")
                if param_abstrtype_link is not None:
                    param_abstrtype = seb.melodyloader.follow_link(
                        seb.melodyobjs[0], param_abstrtype_link
                    )
                    param_name += f":{param_abstrtype.attrib['name']}"
                param_dir = param.get("direction", "IN")
                if param_dir == "RETURN":
                    ret.append(param_name)
                elif param_dir == "EXCEPTION":
                    throw.append(param_name)
                else:
                    params.append(f"{param_dir} {param_name}")

            feat_name += f"({', '.join(params)})"
            if ret or throw:
                feat_name += " : "
            if ret:
                feat_name += f" returns {', '.join(ret)}"
            if throw:
                feat_name += f" throws {', '.join(throw)}"
        else:
            C.LOGGER.warning("Unknown feature type %r", feat_type)
        features[feat_type].append(feat_name)

    box.features = sum(features.values(), list[str]())
    return box


def component_port_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    box = generic_factory(seb)
    try:
        box.styleclass = "CP_" + (seb.melodyobjs[0].attrib["orientation"])
    except KeyError:
        box.styleclass = "CP_UNSET"

    return box


def constraint_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Create the box for a Constraint.

    Constraints are comprised of two parts: A Box and some Edges. The
    Box' label must be extracted from the semantic object, because it
    isn't always easily accessible for the ``generic_factory``.

    See Also
    --------
    capellambse.aird._edge_factories.constraint_factory :
        The accompanying edge factory.
    """
    box = generic_factory(seb)
    label = C.get_spec_text(seb) or seb.melodyobjs[0].attrib.get("name", "")
    if isinstance(label, markupsafe.Markup):
        box.label = label.striptags()
    else:
        box.label = label
    return box


def control_node_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    r"""Differentiate ``ControlNode``\ s based on their KIND."""
    assert seb.styleclass is not None
    kind = seb.melodyobjs[0].get("kind", "OR")
    seb.styleclass = "".join((kind.capitalize(), seb.styleclass))
    return generic_factory(seb)


def enumeration_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Create an Enumeration.

    These work similar to Classes, but use different element tags.
    """
    box = generic_factory(seb)
    box.features = []

    for lit_elm in seb.melodyobjs[0].iterchildren("ownedLiterals"):
        lit_type = lit_elm.attrib[C.ATT_XST]
        if lit_type.split(":")[-1] != "EnumerationLiteral":
            C.LOGGER.warning(
                "Unknown enumeration literal type %r, skipping", lit_type
            )
            continue
        box.features.append(lit_elm.attrib["name"])

    return box


def part_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Resolve the ``Part`` meta-styleclass and creates its Box."""
    dgel_type = seb.diag_element.attrib[C.ATT_XMT]
    if dgel_type in {"diagram:DNodeContainer", "diagram:DNode"}:
        # resolve abstractType reference
        abstract_obj = seb.melodyobjs[0]
        abstract_type = abstract_obj.attrib.get("abstractType")
        if abstract_type is not None:
            abstract_obj = seb.melodyloader.follow_link(
                seb.melodyobjs[0], abstract_type
            )

        seb.styleclass = abstract_obj.attrib[C.ATT_XST].split(":")[-1]
        assert seb.styleclass is not None
        if seb.styleclass.startswith("Physical"):
            seb.styleclass = "".join(
                (
                    "Physical",
                    abstract_obj.get("nature", "Node").capitalize(),
                    seb.styleclass[len("Physical") :],
                )
            )

        if seb.styleclass.endswith("Component"):
            seb.styleclass = "".join(
                (
                    seb.styleclass[: -len("Component")],
                    "Human" * (abstract_obj.get("human") == "true"),
                    ("Component", "Actor")[
                        abstract_obj.get("actor") == "true"
                    ],
                )
            )

        return generic_factory(seb)

    C.LOGGER.error("Unhandled Part type: %r; skipping", dgel_type)
    raise C.SkipObject()


def requirements_box_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Create a Requirement.

    Requirements' text is split in two parts, which have to be joined
    together again for display in diagrams.
    """
    # Only handle the top-level <ownedDiagramElements>,
    # not the nested <ownedElements>.
    if seb.diag_element.tag != "ownedDiagramElements":
        raise C.SkipObject()

    try:
        targetlink = next(seb.diag_element.iterchildren("target"))
        targethref = targetlink.attrib["href"]
    except (StopIteration, KeyError):
        raise C.SkipObject() from None

    seb.melodyobjs[0] = seb.melodyloader.follow_link(targetlink, targethref)
    text = [
        string
        for suffix in ("LongName", "Name", "ChapterName")
        if (string := seb.melodyobjs[0].get("ReqIF" + suffix, ""))
    ]
    if "ReqIFText" in seb.melodyobjs[0].attrib:
        text.append(helpers.repair_html(seb.melodyobjs[0].attrib["ReqIFText"]))

    box = generic_factory(seb, minsize=diagram.Vector2D(0, 0))
    box.features = [f"- {i}" for i in text if i is not None]
    if not (box.floating_labels or box.features):
        sdata_element = seb.data_element.attrib["element"]
        raise ValueError(f"Requirements text is empty for {sdata_element!r}")
    return box


def region_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    r"""Perform special handling for ``Region``\ s.

    *   Adjust the styleclass of Regions to ``StateRegion`` or
        ``ModeRegion``, depending on what the parent box is
    *   Fix the Region's label text
    *   Set the Region's specific minimum / default size (because
        they're different from the usual values)
    """
    parent_elm = seb.diag_element.getparent()
    assert parent_elm is not None
    try:
        parent = seb.target_diagram[parent_elm.attrib["uid"]]
    except KeyError:
        pass
    else:
        seb.styleclass = f'{parent.styleclass or ""}Region'

    box = generic_factory(seb)
    if box.label:
        box.label = f"[{box.label}]"

    box.minsize = (27, 21)
    box.size = diagram.Vector2D(box._size.x or 55, box._size.y or 41)
    return box


def statemode_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Create a State or Mode.

    Unlike other elements, these have their immediate children rendered
    as stacked boxes. The child boxes themselves can in turn have more
    children either in floating or stacked form.

    Additionally, States and Modes can have associated activities. These
    are displayed after the label, separated by a horizontal line.
    """
    xmt = seb.diag_element.get(C.ATT_XMT)
    if xmt == "diagram:DNodeContainer":
        return generic_stacked_factory(seb)
    if xmt == "diagram:DNodeList":
        return statemode_activities_factory(seb)
    C.LOGGER.warning("Unknown State/Mode xmi:type %r", xmt)
    raise C.SkipObject()


def statemode_activities_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Attach the activities to a State or Mode as features."""
    parent_elm = seb.diag_element.getparent()
    assert parent_elm is not None
    parent_id = parent_elm.attrib["uid"]
    try:
        parent = seb.target_diagram[parent_id]
    except KeyError:
        C.LOGGER.error("Cannot find a box with UID %r in diagram", parent_id)
        raise C.SkipObject() from None
    assert isinstance(parent, diagram.Box)

    entry: list[str] = []
    do: list[str] = []
    exit: list[str] = []
    for elm in seb.diag_element.iterchildren("ownedElements"):
        elm_id = elm.get("uid")
        try:
            target_id = next(elm.iterchildren("target")).attrib["href"]
            target = seb.melodyloader[target_id]
            mapping_id = next(elm.iterchildren("actualMapping")).attrib["href"]
        except (KeyError, StopIteration):
            C.LOGGER.error("No usable target or mapping for %r", elm_id)
            continue

        mapping = re.search(
            "@subNodeMappings\\[name=(?P<q>[\"'])(?P<n>.*?)(?P=q)\\]$",
            mapping_id,
        )
        if mapping is not None:
            mapping_name = mapping.group("n")
        try:
            act_list = {
                "MSM_DoActivity": do,
                "MSM_Entry": entry,
                "MSM_Exit": exit,
            }[mapping_name]
        except KeyError:
            C.LOGGER.error("Unknown activity mapping type %r", mapping_name)
            continue
        act_list.append(target.attrib["name"])

    parent.features = list(
        itertools.chain(
            (f" entry / {i}" for i in entry),
            (f" do / {i}" for i in do),
            (f" exit / {i}" for i in exit),
        )
    )
    raise C.SkipObject()


def fcif_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Create a FunctionalChainInvolvementFunction.

    These are special boxes that point to another element via their
    ``involved`` attribute. As there's no relevant information stored in
    the original target attribute, we can simply use the "involved"
    element for constructing the actual box.
    """
    seb.melodyobjs[0] = seb.melodyloader.follow_link(
        seb.melodyobjs[0], seb.melodyobjs[0].attrib["involved"]
    )
    xtype = helpers.xtype_of(seb.melodyobjs[0])
    assert xtype is not None
    if xtype.endswith("Function"):
        seb.styleclass = "Function"
    else:
        seb.styleclass = xtype.split(":")[-1]
    return generic_factory(seb)


def pseudo_symbol_factory(seb: C.SemanticElementBuilder) -> diagram.Box:
    """Create a [Fork|Choice]PseudoState.

    These are boxes that behave like symbols. Capella doesn't store the
    usual workspacePath attribute that references the used image for
    symbols.
    """
    style = next(seb.diag_element.iterchildren("ownedStyle"))
    style.attrib["workspacePath"] = ""
    box = generic_factory(seb)
    box.JSON_TYPE = "box_symbol"
    return box


def _make_portlabel(
    ppos: diagram.Vector2D,
    psize: diagram.Vector2D,
    text: str,
    seb: C.SemanticElementBuilder,
) -> diagram.Box:
    snapsides = {"5001": (1, 0), "5010": (0, 1)}
    try:
        child_elm = next(
            i
            for i in seb.data_element.iterchildren("children")
            if i.get("type") in snapsides
        )
        loc_elm = next(
            i
            for i in child_elm.iterchildren("layoutConstraint")
            if i.get(C.ATT_XMT) == "notation:Location"
        )
    except StopIteration:
        snapside = diagram.Vector2D(1, 0)
    else:
        snapside = (
            diagram.Vector2D(
                int(loc_elm.get("x", "0")), int(loc_elm.get("y", "0"))
            )
            @ snapsides[child_elm.attrib["type"]]
        )
        if snapside == (0, 0):
            snapside = diagram.Vector2D(1, 0)
        else:
            snapside //= snapside.length
    return _make_snapped_floating_label(ppos, psize, text, snapside)


def _make_free_floating_label(
    ppos: diagram.Vector2D,
    psize: diagram.Vector2D,
    text: str,
    seb: C.SemanticElementBuilder,
) -> diagram.Box:
    """Try to construct the label from the real location found in aird file."""
    try:
        child_elm = next(
            i
            for i in seb.data_element.iterchildren("children")
            if i.get("type") == "5002"
        )
        loc_elm = next(
            i
            for i in child_elm.iterchildren("layoutConstraint")
            if i.get(C.ATT_XMT) == "notation:Location"
        )
    except StopIteration:
        return _make_snapped_floating_label(
            ppos, psize, text, diagram.Vector2D(0, -1)
        )
    pos = ppos + (float(loc_elm.get("x", "0")), float(loc_elm.get("y", "0")))
    return diagram.Box(pos, (0, 0), label=text, styleclass="BoxAnnotation")


def _make_snapped_floating_label(
    ppos: diagram.Vector2D,
    psize: diagram.Vector2D,
    text: str,
    snapside: diagram.Vector2D,
) -> diagram.Box:
    if not isinstance(snapside.x, int) or not isinstance(snapside.y, int):
        raise TypeError("snapside must be an int-vector")
    if not (-1 <= snapside.x <= 1 and -1 <= snapside.y <= 1):
        raise ValueError("snapside values must be in interval [-1, 1]")
    if snapside.x and snapside.y:
        raise ValueError("snapside can only have one non-zero value")
    if snapside.x == snapside.y == 0:
        raise ValueError("snapside must have one non-zero value")
    labelbox = diagram.Box(
        (0, 0), (0, 0), label=text, styleclass="BoxAnnotation"
    )
    lsize = labelbox.size
    labelbox.pos = (
        (
            (lambda: ppos.x - (lsize.x - psize.x) / 2),
            (lambda: ppos.x + psize.x + diagram.Box.PORT_OVERHANG),
            (lambda: ppos.x - lsize.x - diagram.Box.PORT_OVERHANG),
        )[snapside.x](),
        (
            (lambda: ppos.y - (lsize.y - psize.y) / 2),
            (lambda: ppos.y + psize.y + diagram.Box.PORT_OVERHANG),
            (lambda: ppos.y - lsize.y - diagram.Box.PORT_OVERHANG),
        )[snapside.y](),
    )
    return labelbox


def _is_collapsed(seb: C.SemanticElementBuilder) -> bool:
    for data_container in seb.data_element.iterchildren():
        if data_container.get("type") == "7002":
            break
    else:
        return False

    for collapsed_container in data_container.iterchildren():
        if collapsed_container.get(C.ATT_XMT) == "notation:DrawerStyle":
            break
    else:
        return False

    return collapsed_container.get("collapsed") == "true"



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import typing as t

import pytest

import capellambse.model.common as c
from capellambse import MelodyModel
from capellambse.model.crosslayer.information import Class

TEMP_PROPERTY_UUID = "2e729284-56e2-4afa-b29e-7fe00d057f80"
CLASS_TYPED_PROP_UUID = "752afd4c-dfa1-4040-baf4-9c5a4ef5b399"
FLOAT_TYPE_UUID = "d65e426c-7df0-43df-aaa4-417ae193176a"


@pytest.mark.parametrize(
    ["name", "super_name", "expected_type"],
    [
        ("SpecialTwist", "Twist", "Class"),
        ("1st Specialization of SuperClass", "SuperClass", "Class"),
        ("SpecialUnion1", "SuperUnion", "Union"),
        ("StatusEnum", "CmdEnum", "Enumeration"),
        ("SpecialCollection1", "SuperCollection", "Collection"),
    ],
)
def test_generalizations(
    model: MelodyModel, name: str, super_name: str, expected_type: str
):
    objects_of_type = model.search(expected_type)
    obj = objects_of_type.by_name(name)
    super_obj = objects_of_type.by_name(super_name)

    obj_type: str = obj.xtype  # type: ignore[assignment]
    assert obj_type.endswith(expected_type)
    assert obj.super == super_obj
    super_xtype: str = super_obj.xtype  # type: ignore[assignment]
    assert super_xtype.endswith(expected_type)
    sub_objects: c.ElementList = super_obj.sub  # type: ignore[assignment]
    assert obj in sub_objects


@pytest.mark.parametrize(
    "uuid,expected_visibility",
    [
        pytest.param(
            "bbc296e1-ed4c-40cf-b37d-c8eb8613228a", "PUBLIC"
        ),  # class
        pytest.param("959b5222-7717-4ee9-bd3a-f8a209899464", "UNSET"),  # class
        pytest.param(
            "d2b4a93c-73ef-4f01-8b59-f86c074ec521", "PACKAGE"
        ),  # class
        pytest.param(
            "ca79bf38-5e82-4104-8c49-e6e16b3748e9", "PROTECTED"
        ),  # class
        pytest.param(TEMP_PROPERTY_UUID, "PUBLIC"),  # property
        pytest.param(
            "3b4915eb-22fc-421d-bf89-07a14d0a2772", "PRIVATE"
        ),  # property
    ],
)
def test_object_visibility(model: MelodyModel, uuid: str, expected_visibility):
    obj = model.by_uuid(uuid)
    assert obj.visibility == expected_visibility


@pytest.mark.parametrize(
    "typed_object_uuid,expected_type_uuid",
    [
        pytest.param(
            "3b4915eb-22fc-421d-bf89-07a14d0a2772",
            "be4e2bfa-2026-4e0b-9a6d-88b5f9c8a3d4",
            id="prop has int type",
        ),
        pytest.param(
            TEMP_PROPERTY_UUID, FLOAT_TYPE_UUID, id="prop has float type"
        ),
        pytest.param(
            CLASS_TYPED_PROP_UUID,
            "d2b4a93c-73ef-4f01-8b59-f86c074ec521",
            id="prop has class type",
        ),
        pytest.param(
            "b6feec5b-3bba-4da9-b9fc-fbd3b72b287d",
            FLOAT_TYPE_UUID,
            id="literal-value has float type",
        ),
    ],
)
def test_object_has_type(
    model: MelodyModel, typed_object_uuid: str, expected_type_uuid: str
):
    obj = model.by_uuid(typed_object_uuid)
    expected_type = model.by_uuid(expected_type_uuid)
    assert obj.type == expected_type


class TestClasses:
    def test_class_owns_stm(self, model: MelodyModel):
        elm: Class = model.by_uuid(
            "959b5222-7717-4ee9-bd3a-f8a209899464"
        )  # type: ignore[assignment]

        assert elm.xtype.endswith("Class")
        assert hasattr(elm, "state_machines")
        assert len(elm.state_machines) == 1

    @pytest.mark.parametrize(
        "uuid,attr_name,expected_value",
        [
            pytest.param(
                "bbc296e1-ed4c-40cf-b37d-c8eb8613228a", "is_abstract", True
            ),
            pytest.param(
                "959b5222-7717-4ee9-bd3a-f8a209899464", "is_abstract", False
            ),
            pytest.param(
                "ca79bf38-5e82-4104-8c49-e6e16b3748e9", "is_final", True
            ),
            pytest.param(
                "959b5222-7717-4ee9-bd3a-f8a209899464", "is_final", False
            ),
            pytest.param(
                "d2b4a93c-73ef-4f01-8b59-f86c074ec521", "is_primitive", True
            ),
            pytest.param(
                "bbc296e1-ed4c-40cf-b37d-c8eb8613228a", "is_primitive", False
            ),
        ],
    )
    def test_class_has_bool_attributes(
        self,
        model: MelodyModel,
        uuid: str,
        attr_name: str,
        expected_value: bool,
    ):
        obj = model.by_uuid(uuid)
        value = getattr(obj, attr_name)
        assert value == expected_value

    @pytest.mark.parametrize(
        "uuid,expected",
        [
            ("bbc296e1-ed4c-40cf-b37d-c8eb8613228a", "PUBLIC"),
            ("ca79bf38-5e82-4104-8c49-e6e16b3748e9", "PROTECTED"),
            ("3b4915eb-22fc-421d-bf89-07a14d0a2772", "PRIVATE"),
            ("d2b4a93c-73ef-4f01-8b59-f86c074ec521", "PACKAGE"),
            ("c371cebb-8021-4a38-8706-4525734de76d", "UNSET"),
        ],
    )
    def test_class_has_visibility(
        self, model: MelodyModel, uuid: str, expected: str
    ) -> None:
        obj = model.by_uuid(uuid)
        assert obj.visibility == expected
        assert not isinstance(obj.visibility, str)

    @pytest.mark.parametrize(
        "uuid,num_of_properties",
        [
            pytest.param("bbc296e1-ed4c-40cf-b37d-c8eb8613228a", 2),
            pytest.param("ca79bf38-5e82-4104-8c49-e6e16b3748e9", 5),
            pytest.param("d2b4a93c-73ef-4f01-8b59-f86c074ec521", 2),
            pytest.param("8164ae8b-36d5-4502-a184-5ec064db4ec3", 0),
        ],
    )
    def test_class_has_properties(
        self, model: MelodyModel, uuid: str, num_of_properties: int
    ):
        obj = model.by_uuid(uuid)
        assert len(obj.properties) == num_of_properties


class TestClassProperty:
    @pytest.mark.parametrize(
        "attr_name",
        [
            pytest.param("is_ordered"),
            pytest.param("is_unique"),
            pytest.param("is_abstract"),
            pytest.param("is_static"),
            pytest.param("is_part_of_key"),
            pytest.param("is_derived"),
            pytest.param("is_read_only"),
        ],
    )
    def test_property_has_bool_attributes(self, model: MelodyModel, attr_name):
        prop_all_false = model.by_uuid(TEMP_PROPERTY_UUID)  # temperature prop
        prop_all_true = model.by_uuid(
            "3b4915eb-22fc-421d-bf89-07a14d0a2772"
        )  # num_of_things prop
        assert hasattr(prop_all_false, attr_name)
        assert getattr(prop_all_false, attr_name) is False
        assert hasattr(prop_all_true, attr_name)
        assert getattr(prop_all_true, attr_name) is True

    @pytest.mark.parametrize(
        "value_attr,expected_val_uuid",
        [
            pytest.param(
                "default_value", "b6feec5b-3bba-4da9-b9fc-fbd3b72b287d"
            ),
            pytest.param("min", "e615cba9-acc4-4a23-ab0c-777c8cabb8f3"),
            pytest.param("max", "d543018f-5f44-4c03-8e2e-875457c8967e"),
            pytest.param("null_value", "9fbdec5d-12c5-4194-b58f-35bff3ee2c9a"),
            pytest.param("min_card", "5546324b-81e8-46db-acee-2f8c1c50afc0"),
            pytest.param("max_card", "581587dd-8301-4bc1-bab4-4f54224d3f6d"),
        ],
    )
    def test_property_has_role_tag_value(
        self, model: MelodyModel, value_attr: str, expected_val_uuid: str
    ):
        obj = model.by_uuid(TEMP_PROPERTY_UUID)
        expected_val = model.by_uuid(expected_val_uuid)
        assert hasattr(obj, value_attr)
        assert getattr(obj, value_attr) == expected_val

    def test_property_has_no_value(self, model: MelodyModel):
        obj = model.by_uuid(CLASS_TYPED_PROP_UUID)
        assert getattr(obj, "min") is None


def test_complex_value(model: MelodyModel):
    cv = model.by_uuid("3a467d68-f53c-4d66-9d32-fe032a8cb2c5")
    value_parts = {
        i.referenced_property.name: i.value.value for i in cv.value_parts
    }
    assert value_parts == {
        "owner": "Harry Potter",
        "core": "Pheonix Feather",
        "wood": "Holly",
    }


@pytest.mark.parametrize(
    "uuid,expected_value,expected_cls",
    [
        pytest.param(
            "1d24c16d-61ad-40b9-9ce0-80e72320e74f",
            "3",
            "LiteralNumericValue",
        ),
        pytest.param(
            "3c6d9df2-1229-4642-aede-a7ac129035c9",
            "Unknown object",
            "LiteralStringValue",
        ),
    ],
)
def test_literal_value_has_value(
    model: MelodyModel, uuid: str, expected_value: t.Any, expected_cls
):
    obj = model.by_uuid(uuid)
    assert obj.value == expected_value
    assert str(obj.xtype).endswith(expected_cls)


def test_literal_value_has_unit(model: MelodyModel):
    obj = model.by_uuid("d543018f-5f44-4c03-8e2e-875457c8967e")
    expected_unit = model.by_uuid("695386d5-6364-4e85-a1c3-a2c489bf0eb2")
    assert obj.unit == expected_unit



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import contextlib
import errno
import logging
import subprocess
from unittest import mock

import pytest

import capellambse
from capellambse.filehandler import git


def test_gitfilehandler_can_read_remote_files_no_revision():
    fh = capellambse.get_filehandler(
        "git+https://github.com/DSD-DBS/py-capellambse.git"
    )
    assert isinstance(fh, git.GitFileHandler)
    assert fh.revision == "refs/heads/master"


def test_gitfilehandler_can_read_remote_files_with_revision():
    fh = capellambse.get_filehandler(
        "git+https://github.com/DSD-DBS/py-capellambse.git",
        revision="gh-pages",
    )

    assert isinstance(fh, git.GitFileHandler)
    assert fh.revision == "refs/heads/gh-pages"


def test_GitFileHandler_locks_repo_during_tasks(monkeypatch, caplog):
    did_ls_files = False

    def mock_run(cmd, *args, encoding="", **kw):
        del args, kw
        nonlocal did_ls_files
        if len(cmd) >= 2 and cmd[1] == "ls-remote":
            assert not did_ls_files
            did_ls_files = True
            data = "0123456789abcdef0123456789abcdef01234567\thello"
            mock_return = mock.Mock()
            mock_return.stdout = data if encoding else data.encode("ascii")
            mock_return.stderr = "" if encoding else b""
            mock_return.returncode = 0
            return mock_return

        assert did_ls_files
        raise FileNotFoundError(errno.ENOENT, "--mocked end of test--")

    did_flock = False

    @contextlib.contextmanager
    def mock_flock(file):
        del file
        nonlocal did_flock
        assert not did_flock
        did_flock = True
        yield

    monkeypatch.setattr(subprocess, "run", mock_run)
    monkeypatch.setattr(capellambse.helpers, "flock", mock_flock)
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    with pytest.raises(FileNotFoundError, match="--mocked end of test--$"):
        capellambse.get_filehandler(
            "git+https://domain.invalid/demo.git", revision="somebranch"
        )

    assert did_ls_files
    assert did_flock



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Facilities for working with a native Capella instance."""
from __future__ import annotations

__all__ = ["ARGS_CMDLINE", "native_capella"]

import collections.abc as cabc
import contextlib
import dataclasses
import errno
import json
import logging
import os
import pathlib
import random
import shutil
import string
import subprocess
import typing as t

import capellambse

_LOGGER = logging.getLogger(__name__)

_Runner = t.Union["_SimpleRunner", "_DockerRunner"]
_COMMON_ARGS: t.Final = (
    "--launcher.suppressErrors",
    "-nosplash",
    "-consolelog",
    "-forceimport",
    "-input",
    "main_model",
)
_ARG_WORKSPACE: t.Final = "-data"
_ARG_PROJECT: t.Final = "-import"

ARGS_CMDLINE: t.Final = (
    "-application",
    "org.polarsys.capella.core.commandline.core",
)


@contextlib.contextmanager
def native_capella(
    model: capellambse.MelodyModel,
    *,
    exe: str | pathlib.Path | None = None,
    docker: str | None = None,
) -> cabc.Iterator[_NativeCapella]:
    """Prepare for running a native Capella instance.

    This context manager prepares a temporary workspace and project
    directory, and provides a function for running a native Capella
    instance with the given arguments.

    It currently supports two ways of running a native Capella instance:

    - Using a native executable, by using the ``exe`` argument to
      specify the path to the executable, or the name of the executable
      if it is in the ``PATH``.
    - Using a Docker image, by specifying the image name with the
      ``docker`` argument.

    Both argument values may be strings containing a ``{VERSION}``
    placeholder, which will be replaced with the Capella version of the
    given model. Note that no substitution will be performed if the
    argument is a :class:`pathlib.Path` object.

    Additionally, to facilitate writing wrapper scripts, the process
    specified by the ``exe`` argument is invoked with the following
    additional environment variables:

    - ``CAPELLA_PROJECT``: The path to the project directory
    - ``CAPELLA_WORKSPACE``: The path to the workspace directory
    - ``CAPELLA_VERSION``: The Capella version of the given model

    The context manager yields a callable that can be used to run the
    native Capella instance. The callable takes the arguments to pass
    to Capella as positional arguments, and returns a
    :class:`subprocess.CompletedProcess` object with the result of the
    execution.

    Some cases may require multiple internal subprocess calls, in which
    case either of them may fail. The caller can therefore not rely on a
    :class:`subprocess.CalledProcessError` to belong to the actual
    Capella execution; it may belong to any of the preparation or
    cleanup calls as well. However, the CompletedProcess object returned
    in case of success always belongs to the Capella process.
    """
    with model._loader.write_tmp_project_dir() as workspace:
        _LOGGER.debug("Project files written to %s", workspace)

        assert model.info.capella_version
        if exe:
            if isinstance(exe, str):
                exe = exe.replace("{VERSION}", model.info.capella_version)
            runner: _Runner = _SimpleRunner(exe, model.info.capella_version)
        elif docker:
            docker = docker.replace("{VERSION}", model.info.capella_version)
            runner = _DockerRunner(docker)
        else:
            assert False

        yield _NativeCapella(workspace, runner)


@dataclasses.dataclass
class _NativeCapella:
    workspace: pathlib.Path
    __runner: _Runner

    def __call__(
        self, *args: str | pathlib.Path
    ) -> subprocess.CompletedProcess:
        try:
            return self.__runner(self.workspace, *args)
        except subprocess.CalledProcessError as err:
            _LOGGER.error(
                "Native Capella failed with code %d\n%s",
                err.returncode,
                err.stdout,
            )
            raise


class _SimpleRunner:
    def __init__(self, path: str | pathlib.Path, capella_version: str) -> None:
        self.capella_version = capella_version
        path = pathlib.Path(path)
        if path.is_absolute():
            self.path = path
        else:
            if path.parent == pathlib.Path("."):
                exe = shutil.which(path.name)
                if exe is None:
                    raise FileNotFoundError(
                        errno.ENOENT, f"Not found in PATH: {path.name}"
                    )
                self.path = pathlib.Path(exe)
            elif path.exists():
                self.path = path.resolve()
            else:
                raise FileNotFoundError(
                    errno.ENOENT, f"File not found: {path.name}"
                )

    def __call__(
        self,
        workspace: pathlib.Path,
        /,
        *args: str | pathlib.Path,
    ) -> subprocess.CompletedProcess:
        env = os.environ | {
            "CAPELLA_PROJECT": str(workspace / "main_model"),
            "CAPELLA_WORKSPACE": str(workspace),
            "CAPELLA_VERSION": self.capella_version,
        }
        cmd: list[str | pathlib.Path] = [
            self.path,
            *_COMMON_ARGS,
            _ARG_WORKSPACE,
            workspace,
            _ARG_PROJECT,
            workspace / "main_model",
            *args,
        ]
        _LOGGER.debug("Running native Capella with command %r", cmd)
        return subprocess.run(
            cmd,
            env=env,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
            text=True,
        )


class _DockerRunner:
    def __init__(self, image: str) -> None:
        self.image = image
        self.username, self.entrypoint = self.get_metadata()

    def __call__(
        self,
        workspace: pathlib.Path,
        /,
        *args: str | pathlib.Path,
    ) -> subprocess.CompletedProcess:
        container_name = "capellambse-" + "".join(
            random.choice(string.ascii_lowercase) for i in range(5)
        )
        with self.start_container(container_name):
            self.copy_files_to_container(
                workspace, "/workspace", container_name
            )

            capella = self.run_capella_in_container(
                container_name, self.entrypoint, *args
            )

            self.copy_files_from_container(
                "/workspace", workspace, container_name
            )

        return capella

    def get_metadata(self) -> tuple[str, list[str]]:
        """Find the username and entrypoint of the Docker image."""
        subprocess.run(["docker", "pull", self.image], check=True)
        proc = subprocess.run(
            ["docker", "inspect", self.image],
            check=True,
            stdout=subprocess.PIPE,
            encoding="utf-8",
        )
        parsed_output = json.loads(proc.stdout)[0]
        ep = parsed_output["Config"]["Entrypoint"]
        if not ep:
            raise ValueError(f"Docker image has no entrypoint: {self.image}")

        username = parsed_output["Config"]["User"]
        return username, ep

    @contextlib.contextmanager
    def start_container(self, name: str) -> cabc.Iterator[None]:
        _LOGGER.debug("Starting container %s from %s", name, self.image)
        with subprocess.Popen(
            [
                "docker",
                "run",
                "--rm",
                "-i",
                "--entrypoint",
                "/bin/sh",
                "--name",
                name,
                self.image,
                "-ec",
                "echo OK; read REPLY || true",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        ) as process:
            assert process.stdin is not None is not process.stdout
            try:
                if process.stdout.readline() != b"OK\n":
                    raise RuntimeError(
                        f"Failed handshake with container image {self.image}"
                    )

                _LOGGER.debug("Started container %s", name)
                yield
            finally:
                _LOGGER.debug("Stopping container %s", name)
                process.communicate(b"\n", timeout=30)

    def run_capella_in_container(
        self,
        container_name: str,
        entrypoint: cabc.Sequence[str],
        *args: str | pathlib.Path,
    ) -> subprocess.CompletedProcess:
        return subprocess.run(
            [
                "docker",
                "exec",
                container_name,
                *entrypoint,
                *_COMMON_ARGS,
                _ARG_WORKSPACE,
                "/workspace",
                _ARG_PROJECT,
                "/workspace/main_model",
                *args,
            ],
            check=True,
        )

    def copy_files_to_container(
        self, directory: pathlib.Path, destination: str, container_name: str
    ) -> None:
        subprocess.check_call(
            [
                "docker",
                "exec",
                "--user=root",
                container_name,
                "rm",
                "-rfv",
                destination,
            ],
        )
        subprocess.check_call(
            [
                "docker",
                "cp",
                f"{directory}/.",
                f"{container_name}:{destination}",
            ],
        )
        subprocess.check_call(
            [
                "docker",
                "exec",
                "--user=root",
                container_name,
                "chown",
                "-R",
                self.username,
                "/workspace",
            ],
        )

    def copy_files_from_container(
        self, directory: str, destination: pathlib.Path, container_name: str
    ) -> None:
        subprocess.check_call(
            [
                "docker",
                "cp",
                f"{container_name}:{directory}/.",
                f"{destination}",
            ],
        )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = [
    "FileHandler",
    "TransactionClosedError",
    "get_filehandler",
]

import logging
import os
import pathlib
import re
import sys
import typing as t
from importlib import metadata

from .abc import *

LOGGER = logging.getLogger(__name__)


def _looks_like_local_path(path: str | os.PathLike) -> bool:
    path = os.fspath(path)
    return bool(
        path.startswith(("/", r"\\")) or re.search(r"^[A-Za-z]:[\\/]", path)
    )


def _looks_like_scp(path: str) -> bool:
    return bool(re.search(r"^(?:\w+@)?[\w.]+:(?:/?[^/].*)?$", path))


def split_protocol(uri: str | os.PathLike) -> tuple[str, str | os.PathLike]:
    """Split the protocol from the URI.

    This function is used to find the name of the file handler for a
    given URI. It takes a URI and returns a tuple of the handler name
    and the potentially modified URI. This function performs the
    following checks and modifications:

    - If the URI either does not contain a protocol, uses the
      ``file://`` protocol, or is a PathLike object, the handler name is
      ``file`` and the URI is converted to a :class:`pathlib.Path`
      object.

    - If the URI is an SCP-style URI (``user@host:path``), the handler
      name is ``git`` and the URI is not modified.

    - If the URI contains nested protocols (e.g. ``git+file://repo``),
      the outermost protocol (i.e. ``git``) is split off and becomes the
      handler name. The rest of the URI (``file://repo``) is returned
      unmodified.
    """
    if _looks_like_local_path(uri):
        return "file", pathlib.Path(uri)
    if isinstance(uri, str) and _looks_like_scp(uri):
        return "git", uri

    uri = os.fspath(uri)
    pattern = r"^(\w+)([+:])"
    prefix_match = re.search(pattern, uri)

    if prefix_match:
        handler_name = prefix_match.group(1)
        if prefix_match.group(2) == "+":
            uri = uri[len(prefix_match.group(0)) :]
        elif handler_name == "file":
            if match := re.match(r"^file://(?:localhost)?/", uri):
                uri = uri[len(match.group(0)) :]
            else:
                raise ValueError(f"Invalid non-local file URI: {uri}")
            uri = pathlib.Path(
                "/" * (not sys.platform.startswith("win")) + uri
            )
    else:
        handler_name = "file"
        uri = pathlib.Path(uri)
    return (handler_name, uri)


def load_entrypoint(handler_name: str) -> type[FileHandler]:
    eps = metadata.entry_points(
        group="capellambse.filehandler", name=handler_name
    )
    if not eps:
        raise ValueError(f"Unknown file handler {handler_name}")
    return next(iter(eps)).load()


def get_filehandler(path: str | os.PathLike, **kwargs: t.Any) -> FileHandler:
    handler_name, path = split_protocol(path)
    handler = load_entrypoint(handler_name)
    return handler(path, **kwargs)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Stylesheet generator for SVG diagrams."""
from __future__ import annotations

import collections.abc as cabc
import itertools
import logging
import operator
import re
import typing as t

from capellambse import diagram

logger = logging.getLogger(__name__)
RE_ELMCLASS = re.compile(r"^([A-Z][a-z_]*)(\.[A-Za-z][A-Za-z0-9_]*)?(:.+)?$")
CUSTOM_STYLE_ATTRS = {"marker-fill"}


class Styling:
    """Container for style attributes of svg objects.

    Notes
    -----
    Attributes containing '-' are only referenceable via getattr() or
    subscripting syntax, due to Python identifier naming rules.
    """

    def __init__(
        self, diagram_class: str | None, class_: str, prefix: str = "", **attr
    ):
        self._diagram_class = diagram_class
        self._class = class_
        self._prefix = prefix

        for key, val in attr.items():
            setattr(self, key, val)

    def __setattr__(self, name: str, value: t.Any) -> None:
        if not name.startswith("_"):
            name = name.replace("_", "-")
        super().__setattr__(name, value)

    def __getattribute__(self, attr: str) -> str:
        if attr in {"marker-start", "marker-end"}:
            defaultstyles = diagram.get_style(self._diagram_class, self._class)
            try:
                value = super().__getattribute__(attr)
            except AttributeError as err:
                try:
                    value = defaultstyles[self._style_name(attr)]
                except KeyError:
                    raise err from None
            try:
                stroke = self.stroke
            except AttributeError:
                stroke = (
                    defaultstyles.get(self._style_name("stroke")) or "#000"
                )
            return f'url("#{value}_{diagram.RGB.fromcss(stroke).tohex()}")'

        return super().__getattribute__(attr)

    def __bool__(self) -> bool:
        try:
            next(iter(self))
        except StopIteration:
            return False
        return True

    def __iter__(self) -> cabc.Iterator[str]:
        defaultstyles = diagram.get_style(self._diagram_class, self._class)
        for attr in ("marker-start", "marker-end"):
            if (
                not getattr(super(), attr, None)
                and self._style_name(attr) in defaultstyles
            ):
                yield attr

        yield from itertools.filterfalse(
            operator.methodcaller("startswith", "_"), dir(self)
        )

    @classmethod
    def _to_css(
        cls, value: float | int | str | diagram.RGB | cabc.Iterable | None
    ) -> float | int | str:
        if isinstance(value, (str, int, float)):
            return value
        elif value is None:
            return "none"
        elif isinstance(value, diagram.RGB):
            return f"#{value.tohex()}"
        elif isinstance(value, cabc.Iterable):
            grad_id = cls._generate_id(
                "CustomGradient",
                value,  # type: ignore[arg-type]  # false-positive
            )
            return f'url("#{grad_id}")'
        raise ValueError(f"Invalid styling value: {value!r}")

    def _to_dict(self) -> dict[str, float | int | str]:
        """Convert this styling to a dictionary.

        The returned dict also includes the built-in default styles for
        the diagram class and object class given to the constructor.

        :meta public:
        """
        styles = diagram.get_style(self._diagram_class, self._class)
        if self._prefix:
            styles = {
                k[len(self._prefix) + 1 :]: v
                for k, v in styles.items()
                if k.startswith(f"{self._prefix}_")
            }
        else:
            styles = {k: v for k, v in styles.items() if "_" not in k}
        styles.update(
            (attr, getattr(self, attr))
            for attr in dir(self)
            if not attr.startswith("_")
        )
        return {k: self._to_css(v) for k, v in styles.items()}

    @staticmethod
    def _generate_id(
        name: str, value: cabc.Iterable[str | diagram.RGB]
    ) -> str:
        """Return unqiue identifier for given css-value."""
        return "_".join(
            itertools.chain(
                (name,),
                (diagram.RGB.fromcss(v).tohex() for v in value),
            ),
        )

    def _style_name(self, attr: str) -> str:
        if not self._prefix:
            return attr

        return "_".join((self._prefix, attr))

    def __contains__(self, obj: t.Any) -> bool:
        return hasattr(self, obj)
